{-# LANGUAGE OverloadedStrings #-}

module Pipes.IRC.Server
  ( startIrcServer
  )
where

import           Control.Concurrent.Async
import           Control.Concurrent.STM
import           Control.Lens                    as L
import           Control.Monad
import           Control.Monad.RWS
import           Data.ByteString.Char8           as BS
import           Data.Map                        as M
import           Data.Set                        as S
import           Network.Socket                  as NS
import           Pipes
import           Pipes.Attoparsec
import           Pipes.Concurrent                as PC
import           Pipes.IRC.Message.Parse
import           Pipes.IRC.Message.Render
import           Pipes.IRC.Message.Types
import           Pipes.IRC.Server.EventHandler
import           Pipes.IRC.Server.Log
import           Pipes.IRC.Server.MessageHandler
import           Pipes.IRC.Server.Types
import           Pipes.Network.TCP               as PN

version :: BS.ByteString
version = "0.1a"

parseMessage :: Producer BS.ByteString IO ()
             -> Producer (Either BS.ByteString IrcMessage) IO ()
parseMessage prod = do
  void $ for (parseMany parseMsgOrLine prod) $ \res ->
    case res of
      (_, Left _)    -> yield $ Left "ERROR Bad Parse"
      (_, Right val) -> yield $ Right val
  return ()

renderMessage :: Pipe IrcMessage BS.ByteString IO ()
renderMessage = forever $ do
  msg <- await
  let output = renderIrcMessage msg
  yield output

filterMsgs :: Pipe (Either BS.ByteString IrcMessage) IrcMessage IO ()
filterMsgs = forever $ do
  cmd <- await
  case cmd of
    Left bs -> liftIO $ BS.putStr $ BS.concat ["BAD COMMAND: ", bs]
    Right c -> do lift $ logMsg c
                  yield c

removeUser :: BS.ByteString -> IrcServer -> IrcServer
removeUser nn ss =
  ss & ircNicks %~ S.delete nn
     & case M.lookup nn (ss ^. ircUsers) of
         Just u -> let ucs = S.elems $ u ^. userChannels in
           (ircUsers %~ M.delete nn) .
           (ircChannels %~ \chmap ->
             Prelude.foldr (M.adjust (chanUsers %~ S.delete nn)) chmap ucs)
         Nothing -> id

addIrcConnection :: ServerState -> IrcConnection -> IO Int
addIrcConnection srv client = do
  let clients = srv ^. ircConnections
      ids = srv ^. ircConnIds
  cid <- atomically $ do
    lastId <- readTVar ids
    let newId = lastId + 1
    writeTVar ids newId
    return newId
  atomically $ modifyTVar' clients $ M.insert cid client
  return cid

delIrcConnection :: ServerState -> Int -> IO ()
delIrcConnection srv cid =
  atomically $ do
    let clients  = srv ^. ircConnections
        srvState = srv ^. ircState
    cs <- readTVar clients
    case M.lookup cid cs of
      -- Connection is unregistered, but has set a nickname
      Just IrcConnection{_reg = Unreg{_rcvdNick = Just nn}} ->
        modifyTVar' srvState $ removeUser nn

      -- Connection is registered
      Just IrcConnection{_reg = RegUser{_regdNick = NickName nn _ _}} ->
        modifyTVar' srvState $ removeUser nn
      _ -> return ()
    modifyTVar' clients $ M.delete cid

cmdHandler :: ServerState -> Int -> Consumer IrcMessage IO ()
cmdHandler srv cid =
  let cReg = Unreg Nothing Nothing Nothing
  in do
    conns <- liftIO $ readTVarIO $ srv ^. ircConnections
    case M.lookup cid conns of
      Just c  -> handle c cReg
      Nothing -> return ()
  where
    handle conn userReg = do
      -- wait for the next command
      nextMsg <- await

      -- run the handler in a transaction
      (newReg, events) <- liftIO $ atomically $ do
        sState <- readTVar $ srv ^. ircState
        let sConf = srv ^. ircConfig
        let cState = ClientState { _clientReg = userReg
                                 , _clientServer = sState
                                 , _clientHost = conn ^. hname
                                 , _clientConn = cid }

        -- run the handler in the IrcMonad, returning new state and events
        let (_, newState, events) =
              runRWS (runIrc $ ircMessageHandler nextMsg) sConf cState

        writeTVar (_ircState srv) $ _clientServer newState
        return (_clientReg newState, events)

      -- handle resulting events
      aliveL <- liftIO $ forM events $ ircEventHandler srv

      -- loop for the next command
      when (and aliveL) $ handle conn newReg

listenHandler :: ServerState -> (Socket, SockAddr) -> IO ()
listenHandler srv (lsock, _) =
  forever $ acceptFork lsock $ \(csock, caddr) -> do
    let sockWriter = toSocket csock
        sockReader = fromSocket csock 4096

    (hName, _) <- getNameInfo [] True False caddr

    (writeEnd, readEnd) <- spawn Unbounded

    let client = IrcConnection
                  { _sock  = csock
                  , _addr  = caddr
                  , _hname = fmap BS.pack hName
                  , _out   = writeEnd
                  , _reg   = Unreg Nothing Nothing Nothing
                  }

    cid <- addIrcConnection srv client

    let handler = cmdHandler srv cid

    r <- async $ runEffect $
           parseMessage sockReader >-> filterMsgs >-> handler
    link r

    w <- async $ runEffect $
           fromInput readEnd >-> renderMessage >-> sockWriter
    link w

    void $ waitEither r w

    delIrcConnection srv cid

mkIrcServer :: IrcConfig -> IO ServerState
mkIrcServer config = do
  let nks = S.empty
      urs = M.empty
      chs = M.empty
      srv = IrcServer nks urs chs version
  tvState <- newTVarIO srv
  tvCns <- newTVarIO M.empty
  tvRef <- newTVarIO 0
  return $ ServerState tvState config tvCns tvRef

startIrcServer :: IrcConfig -> IO (Async ())
startIrcServer config = do
  srv <- mkIrcServer config
  let sHost = srv ^. ircConfig . ircHost
      sPort = srv ^. ircConfig . ircPort
  async $ PN.listen sHost sPort (listenHandler srv)
