{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RecordWildCards   #-}

module Pipes.IRC.Server.MessageHandler
  ( ircMessageHandler )
where

import           Control.Lens
import           Control.Monad.RWS
import qualified Data.ByteString.Char8   as BS
import           Data.Map                as M
import           Data.Maybe
import           Data.Set                as S
import           Pipes.IRC.Message.Types
import           Pipes.IRC.Server.Types



ircMessageHandler :: IrcMessage -> IrcMonad ()
ircMessageHandler msg =
  -- drop messages that have prefixes (until we have Server links)
  when (isNothing $ prefix msg) $ do
    cReg <- use clientReg
    case cReg of
      Unreg   {} -> unregHandler msg
      RegUser {} -> regHandler   msg
    return ()

unregHandler :: IrcMessage -> IrcMonad ()
unregHandler msg@IrcMessage{..} =
  case command of
    Left PASS -> unregPASS msg
    Left NICK -> unregNICK msg
    Left USER -> unregUSER msg
    Left QUIT -> handleQUIT msg
    _         -> return ()

unregPASS :: IrcMessage -> IrcMonad ()
unregPASS IrcMessage{..} =
  if length params < 1
  then tellNumeric err_needmoreparams [":Need more parameters"]
  else do
    clientReg . rcvdPass .= (Just $ head params)
    return ()

validateNick :: BS.ByteString -> IrcMonad Bool
validateNick nickname = do
  nickSet <- use $ clientServer . ircNicks
  if S.member nickname nickSet
   then do
      tellNumeric err_nicknameinuse [nickname, ":Nickname is already in use."]
      return False
   else return True

unregNICK :: IrcMessage -> IrcMonad ()
unregNICK IrcMessage{..} =
  if length params /= 1
  then tellNumeric err_nonicknamegiven [":must supply a nickname"]
  else let nickname = head params in
    validateNick nickname >>= flip when
      (clientReg . rcvdNick .= Just nickname >>
       clientServer . ircNicks %= S.insert nickname >>
       tryRegistration)


unregUSER :: IrcMessage -> IrcMonad ()
unregUSER IrcMessage{..} =
  if length params < 4
  then tellNumeric err_needmoreparams [":need more parameters"]
  else do
    clientReg . rcvdName .= (Just $ head params)
    tryRegistration

useNick :: IrcMonad (Maybe BS.ByteString)
useNick = do
  regState <- use clientReg
  return $ case regState of
    Unreg _ (Just nick) _       -> Just nick
    RegUser (NickName nick _ _) -> Just nick
    _                           -> Nothing

renderQuitMsg :: Maybe BS.ByteString -> IrcMonad BS.ByteString
renderQuitMsg (Just msg) = return $ BS.append "Quit: " msg
renderQuitMsg Nothing    = useNick >>= \nickname ->
  return $ case nickname of
    Just nick -> BS.append "Quit: " nick
    Nothing   -> ""

chanEcho :: IrcMessage -> IrcMonad ()
chanEcho iMsg = do
  mNick <- useNick
  when (isJust mNick) $ do
    let nick = fromJust mNick
    msg <- addUserPrefix iMsg
    mUser <- fmap (M.lookup nick) $ use (clientServer . ircUsers)
    when (isJust mUser) $ do
      let user = fromJust mUser
      let chans = S.elems $ user ^. userChannels
      findReceivers chans >>= fwdMsg msg

doQuit :: IrcMessage -> Maybe BS.ByteString -> IrcMonad ()
doQuit msg quitParamIn = do
  connId <- use clientConn
  quitMsg <- renderQuitMsg quitParamIn
  hostname <- use clientHost
  let hoststr = fromMaybe "unknown hostname" hostname
  let quitParam = BS.concat [ ":Closing Link: ", hoststr
                            , " (", quitMsg, ")"]
  tellCommand ERROR [quitParam]
  chanEcho msg
  tell [Close connId]
  return ()

handleQUIT :: IrcMessage -> IrcMonad ()
handleQUIT msg@IrcMessage{..} =
  doQuit msg $ case params of
    []  -> Nothing
    p:_ -> Just p

mkUser :: NickName -> IrcMonad IrcUser
mkUser nn = do
  conn <- use clientConn
  srvname <- view ircHostName
  return IrcUser { _userNick = nn
                 , _userServerName = srvname
                 , _userModes = S.empty
                 , _userChannels = S.empty
                 , _userConn = conn
                 }

tryRegistration :: IrcMonad ()
tryRegistration = do
  regState <- use clientReg
  hostname <- use clientHost
  case regState of
    Unreg _ (Just nickname) (Just name) -> do
      let nn = NickName nickname (Just name) hostname
      clientReg .= RegUser nn
      newUser <- mkUser nn
      clientServer . ircUsers %= M.insert nickname newUser

      tellWELCOME  nickname
      tellYOURHOST nickname
      tellMOTD     nickname
    _                               -> return ()

addUserPrefix :: IrcMessage -> IrcMonad IrcMessage
addUserPrefix msg = do
  regState <- use clientReg
  return $ case regState ^? regdNick of
    Just nickname -> msg{ prefix = Just . Right $ nickname }
    _             -> msg{ prefix = Nothing }

regHandler :: IrcMessage -> IrcMonad ()
regHandler msg@IrcMessage{..} = do
  pMsg <- addUserPrefix msg
  case command of
    Left PRIVMSG -> handlePRIVMSG pMsg
    Left JOIN    -> return ()
    Left PART    -> return ()
    Left LIST    -> return ()
    Left NICK    -> return ()
    Left QUIT    -> handleQUIT pMsg
    _            -> return ()

handlePRIVMSG :: IrcMessage -> IrcMonad ()
handlePRIVMSG msg@IrcMessage{..} = do
  case params of
    []       -> tellNumeric err_norecipient []
    _:[]     -> tellNumeric err_notexttosend []
    rsp:_:_  -> let rs = Prelude.filter (not . BS.null) $ BS.split ',' rsp
                in findReceivers rs >>= fwdMsg msg
  return ()

channelTargets :: BS.ByteString -> IrcMonad [Int]
channelTargets chname = do
  srv <- use clientServer
  Just mynick <- useNick
  let cUsers chan = S.elems (S.delete mynick $ chan ^. chanUsers)
  let chmap = srv ^. ircChannels
  case M.lookup chname chmap of
    Just chan -> fmap catMaybes $ forM (cUsers chan) userTarget
    _         -> do tellNumeric err_nosuchnick [chname, ":No such nick/channel"]
                    return []

userTarget :: BS.ByteString -> IrcMonad (Maybe Int)
userTarget uname = do
  srv <- use clientServer
  let umap = srv ^. ircUsers
  case M.lookup uname umap of
    Just user -> return . Just $ user ^. userConn
    _         -> do tellNumeric err_nosuchnick [uname, ":No such nick/channel"]
                    return Nothing

findReceivers :: [BS.ByteString] -> IrcMonad [([Int], BS.ByteString)]
findReceivers rcvNames =
  fmap catMaybes $ forM rcvNames $ \name ->
    if BS.head name == '#'
     then do
      cids <- channelTargets name
      return $ case cids of
        [] -> Nothing
        cs -> Just (cs, name)
     else do
      cid <- userTarget name
      return $ case cid of
        Just c  -> Just ([c], name)
        Nothing -> Nothing

fwdMsg :: IrcMessage -> [([Int], BS.ByteString)] -> IrcMonad ()
fwdMsg msg rcvs = forM_ rcvs $ \ (cId, n) -> do
  -- replace multiple targets with the single target we're doing
  let m = msg{ params = n : tail (params msg) }
  tell [Msg m cId]

tellNumeric :: IrcReply -> [IrcParam] -> IrcMonad ()
tellNumeric reply desc = do
  srvname <- view ircHostName
  connId  <- use clientConn
  let msg = IrcMessage (Just $ Left srvname) (Right reply) desc
  tell [Msg msg [connId]]
  return ()

{-
tellPrefixedCommand :: IrcCommand -> [IrcParam] -> IrcMonad ()
tellPrefixedCommand reply desc = do
  srvname <- view ircHostName
  connId <- use clientConn
  let msg = IrcMessage (Just $ Left srvname) (Left reply) desc
  tell [Msg msg [connId]]
-}

tellCommand :: IrcCommand -> [IrcParam] -> IrcMonad ()
tellCommand reply desc = do
  connId <- use clientConn
  let msg = IrcMessage Nothing (Left reply) desc
  tell [Msg msg [connId]]

ppHostPreference :: HostPreference -> BS.ByteString
ppHostPreference hp = case hp of
  HostAny  -> "*"
  HostIPv4 -> "*4"
  HostIPv6 -> "*6"
  Host hn  -> BS.pack hn

ppServiceName :: ServiceName -> BS.ByteString
ppServiceName = BS.pack

tellYOURHOST :: BS.ByteString -> IrcMonad ()
tellYOURHOST nickname = do
  srvname <- view ircHostName
  srvhost <- view ircHost
  srvport <- view ircPort
  version <- use $ clientServer . ircVersion
  let hostStr = ppHostPreference srvhost
      portStr = ppServiceName srvport
  tellNumeric rpl_yourhost
    [ nickname
    , BS.concat [ ":Your host is ", srvname
                , "[", hostStr, "/", portStr, "], "
                , "running version ", version ]
    ]

tellMOTD :: BS.ByteString -> IrcMonad ()
tellMOTD nickname = do
  motd <- view ircMotd
  tellNumeric rpl_motdstart [nickname, ":- Message of the Day -"]
  forM_ motd $ \line ->
    tellNumeric rpl_motd [nickname, ":- " `BS.append` line]
  tellNumeric rpl_endofmotd [nickname, ":End of MOTD"]

tellWELCOME :: BS.ByteString -> IrcMonad ()
tellWELCOME nickname = do
  srvname <- view ircHostName
  tellNumeric rpl_welcome [ nickname
                          , BS.append ":Welcome to IRC on " srvname ]
